package de.fhdw.wtf.common.ast;

import java.util.HashSet;
import java.util.Iterator;

import de.fhdw.wtf.common.ast.type.AtomicType;
import de.fhdw.wtf.common.ast.type.BaseType;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.type.CompositeType;
import de.fhdw.wtf.common.ast.type.ListType;
import de.fhdw.wtf.common.ast.type.MapType;
import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.ast.type.SumType;
import de.fhdw.wtf.common.ast.type.ThrownType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.type.TypeProxy;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.ConstructorOrOperationExceptionVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeVisitor;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.token.Token;

/**
 * This abstract class represents constructors or operations of a class.
 */
public abstract class ConstructorOrOperation extends SyntaxObject {
	
	/**
	 * Represents {@link Type} which will be returned by this constructor or operation.
	 */
	private final Type returnType;
	
	/**
	 * @return Represents the type which includes the {@link ConstructorOrOperation}
	 */
	public Type getReturnType() {
		return this.returnType;
	}
	
	private final ConstructorOrOperationSignature signature;
	
	/**
	 * Creates a new {@link ConstructorOrOperation}.
	 * 
	 * @param parameters
	 *            The parameters of the constructor or operation.
	 * @param containingType
	 *            containingType of the Constructor or Operation
	 * @param returnType
	 * @param firstToken
	 *            firstToken.
	 * @param lastToken
	 *            lastToken.
	 */
	protected ConstructorOrOperation(final String name,
			final ProductType parameters,
			final ClassType containingType,
			final Type returnType,
			final Token firstToken,
			final Token lastToken) {
		super(firstToken, lastToken);
		this.signature = ConstructorOrOperationSignature.create(name, parameters, containingType);
		this.returnType = returnType;
	}
	
	/**
	 * @return Returns the signature of this {@link ConstructorOrOperation}
	 */
	public ConstructorOrOperationSignature getSignature() {
		return this.signature;
	}
	
	@Override
	public abstract boolean equals(final Object o);
	
	@Override
	public abstract int hashCode();
	
	@Override
	public abstract String toString();
	
	/**
	 * Same as getSignature().getParameter()
	 * 
	 * @return the parameter which is assigned to this {@link ConstructorOrOperation}
	 */
	public ProductType getParameters() {
		return this.signature.getParameters();
	}
	
	/**
	 * Same as getSignature().getName()
	 * 
	 * @return the name of this {@link ConstructorOrOperation}
	 */
	public String getName() {
		return this.signature.getName();
	}
	
	/**
	 * Same as getSignature().getContainingType()
	 * 
	 * @return the containingType which is assigned to this {@link ConstructorOrOperation}
	 */
	public ClassType getContainingType() {
		return this.signature.getContainingType();
	}
	
	public HashSet<Type> getThrownExceptions() {
		final HashSet<Type> outputLst = new HashSet<Type>();
		
		this.getReturnType().accept(new TypeVisitor() {
			
			@Override
			public void handle(final TypeProxy typeProxy) {
				// TODO Auto-generated method stub
				
			}
			
			@Override
			public void handle(final CompositeType compositeType) {
				compositeType.accept(new CompositeTypeVisitor() {
					
					@Override
					public void handle(final ThrownType thrownType) {
						outputLst.add(thrownType);
						
					}
					
					@Override
					public void handle(final SumType sum) {
						
						final Iterator<ThrownType> itera =
								ConstructorOrOperation.this.getThrownTypeHelperSum(sum).iterator();
						while (itera.hasNext()) {
						}
					}
					
					@Override
					public void handle(final ProductType product) {
						final Iterator<ThrownType> itera =
								ConstructorOrOperation.this.getThrownTypeHelperPro(product).iterator();
						while (itera.hasNext()) {
							outputLst.add(itera.next());
						}
					}
					
					@Override
					public void handle(final MapType map) {
						// nothing to do here
						
					}
					
					@Override
					public void handle(final ListType list) {
						// nothing to do here;
						
					}
				});
				
			}
			
			@Override
			public void handle(final AtomicType atomicType) {
				atomicType.accept(new AtomicTypeVisitor() {
					
					@Override
					public void handle(final ClassType clazz) {
						if (clazz.isThrownType()) {
							outputLst.add(clazz);
						}
						
					}
					
					@Override
					public void handle(final BaseType baseType) {
						// nothing To Do Here!
						
					}
				});
				
			}
		});
		
		return outputLst;
	}
	
	private HashSet<ThrownType> getThrownTypeHelperSum(final SumType input) {
		final HashSet<ThrownType> output = new HashSet<ThrownType>();
		return output;
	}
	
	private HashSet<ThrownType> getThrownTypeHelperPro(final ProductType input) {
		final HashSet<ThrownType> output = new HashSet<ThrownType>();
		return output;
	}
	
	/**
	 * @param constructorAndOperationVisitor
	 * @throws TaskException
	 */
	public abstract void accept(final ConstructorOrOperationExceptionVisitor<TaskException> constructorAndOperationVisitor)
			throws TaskException;
}
