package de.fhdw.wtf.common.ast.type;

import java.util.ArrayList;
import java.util.List;

import de.fhdw.wtf.common.ast.DatabaseIDState;
import de.fhdw.wtf.common.ast.SyntaxObject;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.TypeVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturnException;
import de.fhdw.wtf.common.exception.parser.SumIsAnythingException;
import de.fhdw.wtf.common.token.Token;

/**
 * This class represents the Type of an <Attribute> or a Parameter of an <Operation>. A type can be a {@link TypeProxy},
 * a {@link AtomicType} or {@link CompositeType}.
 */
public abstract class Type extends SyntaxObject implements Comparable<Type> {
	
	// compareTo ist wie folgt programmiert: Class < Product < Sum < Map < List
	// < Basetype < TypeProxy
	
	/**
	 * Prototype is a standardized Type of this.
	 */
	private Type prototype;
	
	/**
	 * The superTypes of the {@link Type}.
	 */
	private final List<Type> superTypes;
	
	/**
	 * Creates a {@link Type}.
	 * 
	 * @param firstToken
	 *            The token starting the type definition.
	 * @param lastToken
	 *            The first token after the type definition.
	 */
	protected Type(final Token firstToken, final Token lastToken) {
		super(firstToken, lastToken);
		this.superTypes = new ArrayList<>();
	}
	
	/**
	 * Creates a {@link Type}.
	 * 
	 * @param firstToken
	 *            The token starting the type definition.
	 * @param lastToken
	 *            The first token after the type definition.
	 * @param superTypes
	 *            A collection of super types.
	 */
	protected Type(final Token firstToken, final Token lastToken, final List<Type> superTypes) {
		super(firstToken, lastToken);
		this.superTypes = superTypes;
	}
	
	/**
	 * Returns a string representation of this type. Atomic types return their name. Composite types return a string
	 * describing their structure. In contrast to {@link #toString()}, this operation does not expose the internals of a
	 * class type.
	 * 
	 * @return A string representation of this type.
	 */
	public abstract String getTypeString();
	
	/**
	 * Returns the database ID of this type.
	 * 
	 * @return The database ID of this type.
	 */
	public abstract DatabaseIDState getTypeId();
	
	/**
	 * Sets the database ID of this type.
	 * 
	 * @param typeId
	 *            The new database ID.
	 */
	public abstract void setTypeId(final DatabaseIDState typeId);
	
	@Override
	public abstract boolean equals(Object obj);
	
	@Override
	public abstract int hashCode();
	
	/**
	 * Accept-Method from Visitor-Pattern.
	 * 
	 * @param v
	 *            visitor
	 */
	public abstract void accept(TypeVisitor v);
	
	/**
	 * Accept-Method from Visitor-Pattern.
	 * 
	 * @param <X>
	 *            ReturnType
	 * @param v
	 *            visitor
	 * @return X
	 */
	public abstract <X> X accept(TypeVisitorReturn<X> v);
	
	/**
	 * Accept-Method from Visitor-Pattern.
	 * 
	 * @param <X>
	 *            ExceptionType
	 * @param v
	 *            visitor
	 * @throws X
	 *             Exception
	 * 
	 */
	public abstract <X extends Exception> void accept(TypeVisitorException<X> v) throws X;
	
	/**
	 * Accept-Method from Visitor-Pattern.
	 * 
	 * @param <X>
	 *            ReturnType
	 * @param <Y>
	 *            ExceptionType
	 * @param v
	 *            visitor
	 * @return X
	 * @throws Y
	 *             Exception
	 */
	public abstract <X, Y extends Exception> X accept(TypeVisitorReturnException<X, Y> v) throws Y;
	
	/**
	 * Returns true if passed type is more special than (i.e. a subType of) this type. Otherwise false. <br/>
	 * <br/>
	 * It return only a right result, if the {@link de.fhdw.wtf.walker.tasks.CyclicInheritanceCheck} and the
	 * {@link de.fhdw.wtf.walker.tasks.SubtypesFillTask} runs already successful.
	 * 
	 * @param type
	 *            type
	 * @return result
	 */
	public abstract boolean isSupertypeOf(Type type);
	
	/**
	 * Set the prototype. A prototype is a standardized representation of this Type.
	 * 
	 * @param prototype
	 *            prototype
	 */
	public void setPrototype(final Type prototype) {
		this.prototype = prototype;
	}
	
	/**
	 * If it set from the {@link de.fhdw.wtf.walker.tasks.PrototypesTask}, then this returned a standardized
	 * representation of this Type. A Prototype returned itself as prototype.
	 * 
	 * @return prototype
	 */
	public Type getPrototype() {
		return this.prototype;
	}
	
	/**
	 * Returns the collection of superTypes of this.
	 * 
	 * @return superTypes
	 */
	public List<Type> getSuperTypes() {
		return this.superTypes;
	}
	
	/**
	 * Returns a new standardized object of the concrete type.
	 * 
	 * @throws SumIsAnythingException
	 *             If the method found Anything as Element in a SumType, this Exception will be thrown.
	 * @return standardized type
	 */
	public abstract Type standardize() throws SumIsAnythingException;
	
	/**
	 * Check if the type is a ThrownType.
	 * 
	 * @return true if the type is a ThrownType, else it returns false
	 */
	public abstract boolean isThrownType();
	
	/**
	 * returns true, if <<this>> is more special than <<typeToCompare>>
	 * 
	 * @param another
	 *            Type, which can be more special or not
	 * @return
	 */
	public boolean isMoreOrEqualSpecialThan(final Type typeToCompare) {
		// return (this.getSuperTypes().contains(typeToCompare));
		
		return this.accept(new TypeVisitorReturn<Boolean>() {
			
			@Override
			public Boolean handle(final AtomicType atomicType) {
				return typeToCompare.accept(new TypeVisitorReturn<Boolean>() {
					
					@Override
					public Boolean handle(final AtomicType atomicType) {
						return (Type.this.getSuperTypes().contains(typeToCompare));
					}
					
					@Override
					public Boolean handle(final CompositeType compositeType) {
						return compositeType.accept(new CompositeTypeVisitorReturn<Boolean>() {
							
							@Override
							public Boolean handle(final ListType list) {
								return false;
							}
							
							@Override
							public Boolean handle(final MapType map) {
								return false;
							}
							
							@Override
							public Boolean handle(final ProductType product) {
								if (product.getElements().contains(Type.this)) {
									return true;
								}
								return false;
							}
							
							@Override
							public Boolean handle(final SumType sum) {
								if (sum.getElements().contains(Type.this)) {
									return true;
								}
								return false;
							}
							
							@Override
							public Boolean handle(final ThrownType thrownType) {
								// TODO Auto-generated method stub
								return null;
							}
						});
					}
					
					@Override
					public Boolean handle(final TypeProxy typeProxy) {
						return (Type.this.getSuperTypes().contains(typeProxy.getTarget()));
					}
					
				});
			}
			
			@Override
			public Boolean handle(final CompositeType compositeType) {
				return compositeType.accept(new CompositeTypeVisitorReturn<Boolean>() {
					
					@Override
					public Boolean handle(final ListType list) {
						if (list.equals(typeToCompare)) {
							return true;
						}
						if (list.getSuperTypes().contains(typeToCompare)) {
							return true;
						}
						return false;
					}
					
					@Override
					public Boolean handle(final MapType map) {
						if (map.equals(typeToCompare)) {
							return true;
						}
						if (map.getSuperTypes().contains(typeToCompare)) {
							return true;
						}
						return false;
					}
					
					@Override
					public Boolean handle(final ProductType product) {
						if (product.equals(typeToCompare)) {
							return true;
						}
						if (product.getSuperTypes().contains(typeToCompare)) {
							return true;
						}
						return false;
					}
					
					@Override
					public Boolean handle(final SumType sum) {
						final java.util.Iterator<Type> itr = sum.getElements().iterator();
						while (itr.hasNext()) {
							return itr.next().isMoreOrEqualSpecialThan(typeToCompare);
						}
						return false;
					}
					
					@Override
					public Boolean handle(final ThrownType thrownType) {
						return false;
					}
				});
			}
			
			@Override
			public Boolean handle(final TypeProxy typeProxy) {
				return (Type.this.getSuperTypes().contains(typeProxy.getTarget()));
				
			}
			
		});
	}
}
