package de.fhdw.wtf.dsl.scanner.common;

import java.util.List;

import de.fhdw.wtf.common.stream.ScannerInputStream;
import de.fhdw.wtf.common.stream.TokenStream;
import de.fhdw.wtf.common.token.EndToken;
import de.fhdw.wtf.common.token.Position;

/**
 * <h1>Scanner</h1> The {@link Scanner} can be used to transform <i>WTF-code</i> in the form of a
 * {@link ScannerInputStream} into the corresponding {@link TokenStream}.
 * 
 * @author hf(w|i|t)411
 */
public abstract class Scanner {
	
	/**
	 * Width of the tabulator.
	 */
	private final int tabulatorWidth;
	
	/**
	 * If this character appears in the input stream, the position will be adjusted to show a new line.
	 */
	private static final char LINE_FEED = '\n';
	
	/**
	 * If this character appears in the input stream, the tabulatorWidth will be add to the position.
	 */
	private static final char TAB = '\t';
	
	/**
	 * Constructor for {@link Scanner}.
	 * 
	 * @param tabulatorWidth
	 *            Width of TAB (editor-setting) for calculation of token column positions
	 */
	protected Scanner(final int tabulatorWidth) {
		this.tabulatorWidth = tabulatorWidth;
	}
	
	/**
	 * Returns the delta-function of the Scanner-automata by defining all transitions ({@link Delta}s).
	 * 
	 * @return delta-function
	 */
	protected abstract List<Delta> getDeltas();
	
	/**
	 * Returns the first {@link ScannerState}.
	 * 
	 * @return first {@link ScannerState}
	 */
	protected abstract ScannerState getStartState();
	
	/**
	 * Returns the {@link ScannerState} witch will be used if no {@link Delta} has matched at
	 * {@link #step(ScannerState, Character, Position, TokenStream)} e.g. an invalid state.
	 * 
	 * @param position
	 *            position
	 * @return Default-State
	 */
	protected abstract ScannerState getDefaultState(Position position);
	
	/**
	 * Reads the given {@link ScannerInputStream}, scans regular language tokens and writes the corresponding Tokens
	 * into the given {@link TokenStream}.
	 * 
	 * @param inputStream
	 *            where characters are read from
	 * @param outputStream
	 *            where created Tokens are added
	 */
	public void scan(final ScannerInputStream inputStream, final TokenStream outputStream) {
		ScannerState state = this.getStartState();
		final String filePath = "";
		// TODO Fix filePath when imports are allowed in grammar (Bug #769)
		int lineNumber = 1;
		int positionInFile = 0;
		int positionInLine = 1;
		
		while (inputStream.hasNext()) {
			final char current = inputStream.next();
			final Position position = Position.create(filePath, lineNumber, positionInLine, positionInFile);
			state = this.step(state, current, position, outputStream);
			state.process(current, outputStream, position);
			
			if (current == TAB) {
				final int spacesToInsert = this.tabulatorWidth - ((positionInLine - 1) % this.tabulatorWidth);
				positionInLine += spacesToInsert;
			} else {
				positionInLine++;
			}
			
			positionInFile++;
			
			// Zeilenumbruch
			if (current == LINE_FEED) {
				lineNumber++;
				positionInLine = 1;
			}
		}
		state.finish(outputStream);
		outputStream.add(EndToken.create(Position.create(filePath, lineNumber, positionInLine, positionInFile)));
	}
	
	/**
	 * Returns the next {@link ScannerState} and finishes the last {@link ScannerState}.
	 * 
	 * @param currentState
	 *            the current {@link ScannerState}
	 * @param c
	 *            the current input {@link Character}
	 * @param position
	 *            the current {@link Position}
	 * @param outputStream
	 *            {@link TokenStream}
	 * @return the next {@link ScannerState}
	 */
	private ScannerState step(final ScannerState currentState,
			final Character c,
			final Position position,
			final TokenStream outputStream) {
		for (final Delta d : this.getDeltas()) {
			if (d.match(currentState, c)) {
				final ScannerState nextState = d.getNextScannerState(position, currentState);
				if (!nextState.equals(currentState)) {
					currentState.finish(outputStream);
				}
				return nextState;
			}
		}
		currentState.finish(outputStream);
		return this.getDefaultState(position);
	}
}
