package de.fhdw.wtf.walker.tasks;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.Constructor;
import de.fhdw.wtf.common.ast.ConstructorByReferenceState;
import de.fhdw.wtf.common.ast.ConstructorByTypeAndSignatureState;
import de.fhdw.wtf.common.ast.ConstructorInvalidState;
import de.fhdw.wtf.common.ast.ConstructorOrOperation;
import de.fhdw.wtf.common.ast.ConstructorReference;
import de.fhdw.wtf.common.ast.Group;
import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.Operation;
import de.fhdw.wtf.common.ast.Tuple;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.visitor.ConstructorOrOperationExceptionVisitor;
import de.fhdw.wtf.common.ast.visitor.ConstructorReferenceStateVisitorReturn;
import de.fhdw.wtf.common.exception.referencer.ConstructorReferenceMismatchException;
import de.fhdw.wtf.common.exception.referencer.ConstructorReferencerExceptions;
import de.fhdw.wtf.common.exception.referencer.CyclicConstructorCallException;
import de.fhdw.wtf.common.exception.referencer.DuplicateConstructorException;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.TaskExecutor;

/**
 * This referencer tries to find the real {@link Constructor} instances that are referenced by the
 * {@link ConstructorReference} instances. The found {@link ConstructorReference} instances in the superConstructor
 * collections of all defined {@link Constructor} instances are being replaced by matching {@link Constructor}
 * instances. This task shall not be performed before the task {@link TypeReferencer}.
 *
 */
public class ConstructorReferencer extends Referencer {
	
	/**
	 * The minimum of {@link Constructor} instances that are searched to replace a {@link ConstructorReference}.
	 * 
	 */
	private static final int MINIMUM_OF_CONSTRUCTOR_INSTANCES_NEEDED_TO_MATCH_A_REFERENCE = 1;
	
	/**
	 * The maximum of {@link Constructor} instances that are searched to replace a {@link ConstructorReference}. The
	 * transgression of that value can be seen as exceptional behavior (Wrong wtf-model).
	 */
	private static final int MAXIMUM_OF_CONSTRUCTOR_INSTANCES_ALLOWED_TO_MATCH_A_REFERENCE = 1;
	
	/**
	 * COntains all {@link ClassType} instances that contain at least one {@link Constructor} instance (which should be
	 * all {@link ClassType} instances).
	 */
	private final Collection<ClassType> constructorOwners;
	/**
	 * Contains tuples of {@link ConstructorReference} instances and the {@link Constructor} instances which contain
	 * them.
	 */
	private final Collection<Tuple<ConstructorReference, Constructor>> constructorReferences;
	/**
	 * Contains all exception that might appear when this task is being performed. This field should be read to
	 * determine whether an exception happened or not.
	 */
	private final Collection<TaskException> exceptions;
	
	/**
	 * Instantiates a new {@link ConstructorReferencer}.
	 * 
	 * @param model
	 *            the model with the {@link ConstructorReference} instances to replace.
	 * @param taskmanager
	 *            the manager which performs this task.
	 */
	public ConstructorReferencer(final Model model, final TaskExecutor taskmanager) {
		super(model, taskmanager);
		this.constructorOwners = new LinkedList<>();
		this.constructorReferences = new LinkedList<>();
		this.exceptions = new LinkedList<>();
	}
	
	/**
	 * Creates a {@link ConstructorReferencer}-Object.
	 * 
	 * @param inputModel
	 *            Model
	 * @param taskmanager
	 *            Taskmanager
	 * @return The Referencer-Object
	 */
	public static ConstructorReferencer create(final Model inputModel, final TaskExecutor taskmanager) {
		return new ConstructorReferencer(inputModel, taskmanager);
	}
	
	@Override
	public void handleClass(final ClassType c) throws TaskException {
		if (c.getConstructors().isEmpty()) {
			this.exceptions.add(new TaskException("No constructor for class " + c.getTypeName()
					+ " has bee defined!\nPlease define a constructor explicitely."));
		}
		this.constructorOwners.add(c);
	}
	
	@Override
	public void handleGroup(final Group g) throws TaskException {
		// Nothing to do.
	}
	
	@Override
	public void handleAttribute(final Attribute a, final ClassType owner) throws TaskException {
		// Nothing to do.
	}
	
	@Override
	public void handleConstructorOrOperation(final ConstructorOrOperation coo, final ClassType owner)
			throws TaskException {
		coo.accept(new ConstructorOrOperationExceptionVisitor<TaskException>() {
			
			@Override
			public void handleOperation(final Operation operation) throws TaskException {
				// Nothing to do.
				
			}
			
			@Override
			public void handleConstructor(final Constructor constructor) throws TaskException {
				final Iterator<ConstructorReference> referencedConstructors =
						constructor.getSuperConstructors().iterator();
				while (referencedConstructors.hasNext()) {
					final ConstructorReference referencedConstructor = referencedConstructors.next();
					ConstructorReferencer.this.constructorReferences
							.add(new Tuple<>(referencedConstructor, constructor));
				}
				
			}
		});
		
	}
	
	@Override
	public void finalizeTask() throws TaskException {
		final Iterator<Tuple<ConstructorReference, Constructor>> references = this.constructorReferences.iterator();
		while (references.hasNext()) {
			final Tuple<ConstructorReference, Constructor> currentReferenceTuple = references.next();
			final ConstructorReference currentReference = currentReferenceTuple.getFirst();
			final Constructor callingConstructor = currentReferenceTuple.getSecond();
			final Iterator<ClassType> constructorOwnersIterator = this.constructorOwners.iterator();
			boolean matchFound = false;
			while (!matchFound && constructorOwnersIterator.hasNext()) {
				final ClassType currentConstructorOwner = constructorOwnersIterator.next();
				matchFound =
						this.replaceReferenceWithRealConstructorIfMatchFound(
								currentConstructorOwner,
								callingConstructor,
								currentReference);
			}
			if (!matchFound) {
				this.exceptions.add(ConstructorReferenceMismatchException.create(currentReference.getFirstToken()));
			}
		}
		if (!this.exceptions.isEmpty()) {
			throw ConstructorReferencerExceptions.create(this.exceptions);
		}
	}
	
	/**
	 * Tries to find exactly one match between reference, if it has only a type and the list of parameters, and one of
	 * constructorOwners constructors and replaces reference in callingConstructor with the matched constructor. Returns
	 * true if the above applies. Adds a DuplicateConstructorException to exceptions if more than one match is being
	 * found.
	 * 
	 * @param constructorOwner
	 *            a class with a list of constructors to match
	 * @param callingConstructor
	 *            the constructor that references another constructor
	 * @param reference
	 *            the reference to replace with a real constructor.
	 * @return true if one or more matching constructors have been found (Regardless of the fact that this "causes" an
	 *         exception). Otherwise false.
	 */
	private boolean replaceReferenceWithRealConstructorIfMatchFound(final ClassType constructorOwner,
			final Constructor callingConstructor,
			final ConstructorReference reference) {
		
		return reference.getState().accept(new ConstructorReferenceStateVisitorReturn<Boolean>() {
			
			@Override
			public Boolean handle(final ConstructorByTypeAndSignatureState byName) {
				boolean matchFound = false;
				final Collection<Constructor> matchedConstuctors = constructorOwner.getMatchingConstructors(byName);
				matchFound = !matchedConstuctors.isEmpty();
				if (matchedConstuctors.size() == ConstructorReferencer.MINIMUM_OF_CONSTRUCTOR_INSTANCES_NEEDED_TO_MATCH_A_REFERENCE
						&& !ConstructorReferencer.this.constructorCycle(callingConstructor, matchedConstuctors)) {
					reference.setState(ConstructorByReferenceState.create(
							matchedConstuctors.iterator().next(),
							constructorOwner));
				} else if (matchedConstuctors.size() > ConstructorReferencer.MAXIMUM_OF_CONSTRUCTOR_INSTANCES_ALLOWED_TO_MATCH_A_REFERENCE) {
					ConstructorReferencer.this.exceptions.add(DuplicateConstructorException.create(
							matchedConstuctors,
							callingConstructor.getFirstToken()));
				}
				return matchFound;
			}
			
			@Override
			public Boolean handle(final ConstructorInvalidState invalid) {
				// nothing to do
				return false;
			}
			
			@Override
			public Boolean handle(final ConstructorByReferenceState byReference) {
				// nothing to do
				return false;
			}
		});
		
	}
	
	/**
	 * Returns true if callingConstructor is part of matchedConstructors. Adds a CyclicConstructorCallException to
	 * exceptions and returns false otherwise.
	 * 
	 * @param callingConstructor
	 *            the constructor search for within matchedConstructors.
	 * @param matchedConstuctors
	 *            matchedConstructors.
	 * @return a boolean as stated above.
	 */
	private boolean constructorCycle(final Constructor callingConstructor,
			final Collection<Constructor> matchedConstuctors) {
		final boolean result = matchedConstuctors.contains(callingConstructor);
		if (result) {
			this.exceptions.add(CyclicConstructorCallException.create(
					callingConstructor,
					callingConstructor.getFirstToken()));
		}
		return result;
	}
	
	@Override
	public void beginTask() throws TaskException {
		// Nothing to do.
	}
	
}
