package de.fhdw.wtf.walker.tasks;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.type.AtomicType;
import de.fhdw.wtf.common.ast.type.CompositeType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.type.TypeProxy;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.task.DependencyTask;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.common.task.result.OKTaskResult;
import de.fhdw.wtf.common.task.result.TaskResult;
import de.fhdw.wtf.walker.walker.HelperUtils;

/**
 * This task clean by every PROTOTYPE the List of Supertypes.<br>
 * - remove transitive inheritances <br>
 * - remove duplicate inheritance
 * 
 * This Checker must run after the {@link SupertypesFillTask}.
 */
public final class SupertypesCleanInheritanceTask extends DependencyTask {
	
	/**
	 * All Sums of the model.
	 */
	private final Set<Type> allTypesOfModel;
	
	/**
	 * The model.
	 */
	private final Model model;
	
	/**
	 * supertypesFillTask.
	 */
	private final SupertypesFillTask supertypesFillTask;
	
	/**
	 * Constructor of {@link SupertypesCleanInheritanceTask}.
	 * 
	 * @param m
	 *            the model
	 * @param taskmanager
	 *            taskmanager
	 * @param supertypesFillTask
	 *            SupertypesFillTask, is need to get the all types of the model.
	 * 
	 */
	public SupertypesCleanInheritanceTask(final Model m,
			final TaskExecutor taskmanager,
			final SupertypesFillTask supertypesFillTask) {
		super(taskmanager);
		this.model = m;
		this.allTypesOfModel = new HashSet<>();
		this.supertypesFillTask = supertypesFillTask;
	}
	
	/**
	 * Creates a {@link SupertypesCleanInheritanceTask}-Object.
	 * 
	 * @param m
	 *            the model
	 * @param taskmanager
	 *            taskmanager
	 * @param sumSupertypesFillTask
	 *            SumSupertypesFillTask, is need to get the sums.
	 * 
	 * @return The {@link SupertypesCleanInheritanceTask}-Object.
	 */
	public static SupertypesCleanInheritanceTask create(final Model m,
			final TaskExecutor taskmanager,
			final SupertypesFillTask sumSupertypesFillTask) {
		return new SupertypesCleanInheritanceTask(m, taskmanager, sumSupertypesFillTask);
	}
	
	@Override
	public TaskResult doWork() {
		this.allTypesOfModel.addAll(this.supertypesFillTask.getAllTypesOfModel());
		
		for (final Type current : this.getAllTypesOfModel()) {
			// Delete duplicate Inheritance
			final Collection<Type> helperCollection = new ArrayList<>();
			
			for (final Iterator<Type> iterator = current.getSuperTypes().iterator(); iterator.hasNext();) {
				final Type supertype = iterator.next();
				if (helperCollection.contains(supertype)) {
					iterator.remove();
				} else {
					helperCollection.add(supertype);
				}
			}
			
			// Clean Inheritance: transitive Inheritance
			final Collection<Type> unnecessaryTransitiveInheritance = this.getUnnecessaryTransitiveInheritance(current);
			current.getSuperTypes().removeAll(unnecessaryTransitiveInheritance);
		}
		
		return new OKTaskResult();
	}
	
	/**
	 * Returns the list of unnecessary Inheritance, because of transitive Inheritance.
	 * 
	 * @param type
	 *            Type
	 * @return unnecessary Inheritance
	 */
	private Collection<Type> getUnnecessaryTransitiveInheritance(final Type type) {
		final Collection<Type> superTypes = type.getSuperTypes();
		final Collection<Type> unnecessaryTransitiveInheritance = new ArrayList<>();
		for (final Type superType : superTypes) {
			for (final Type superTypeOfSuperType : HelperUtils.getReferencedType(superType).getSuperTypes()) {
				unnecessaryTransitiveInheritance.addAll(this.getUnnecessaryTransitiveInheritance(
						superTypeOfSuperType,
						superTypes));
			}
		}
		return unnecessaryTransitiveInheritance;
	}
	
	/**
	 * Helper-Method of getUnnecessaryTransitiveInheritance(Type).
	 * 
	 * @param superType
	 *            superType
	 * @param superTypes
	 *            superTypes of original type
	 * @return unnecessary Inheritance
	 */
	Collection<Type> getUnnecessaryTransitiveInheritance(final Type superType, final Collection<Type> superTypes) {
		return superType.accept(new TypeVisitorReturn<Collection<Type>>() {
			
			@Override
			public Collection<Type> handle(final AtomicType atomicType) {
				// Only as typeProxy in supertypes so here we call the method with the supertypes directly
				final Collection<Type> unnecessaryTransitiveInheritance = new ArrayList<>();
				SupertypesCleanInheritanceTask.this.checkUnnecessaryInheritanceInTheSuperTypesOfTheSuperType(
						atomicType,
						superTypes,
						unnecessaryTransitiveInheritance);
				
				return unnecessaryTransitiveInheritance;
			}
			
			@Override
			public Collection<Type> handle(final CompositeType compositeType) {
				final Collection<Type> unnecessaryTransitiveInheritance = new ArrayList<>();
				// If the type itself contains in supertype it is unnessary
				SupertypesCleanInheritanceTask.this.addTypeToUnnecessaryInheritanceIfSupertypesContainsIt(
						compositeType,
						superTypes,
						unnecessaryTransitiveInheritance);
				// Now we look in the supertypes
				SupertypesCleanInheritanceTask.this.checkUnnecessaryInheritanceInTheSuperTypesOfTheSuperType(
						compositeType,
						superTypes,
						unnecessaryTransitiveInheritance);
				
				return unnecessaryTransitiveInheritance;
			}
			
			@Override
			public Collection<Type> handle(final TypeProxy typeProxy) {
				final Collection<Type> unnecessaryTransitiveInheritance = new ArrayList<>();
				// Anything and AtomicTypes are as TypeProxy in the list of supertypes
				SupertypesCleanInheritanceTask.this.addTypeToUnnecessaryInheritanceIfSupertypesContainsIt(
						typeProxy,
						superTypes,
						unnecessaryTransitiveInheritance);
				
				// Look in the typeProxy to get the supertypes-Info
				unnecessaryTransitiveInheritance.addAll(SupertypesCleanInheritanceTask.this
						.getUnnecessaryTransitiveInheritance(HelperUtils.getTargetType(typeProxy), superTypes));
				
				return unnecessaryTransitiveInheritance;
			}
		});
	}
	
	/**
	 * Add the type to list of unnecessaryTransitiveInheritance, if the type is in the list of supertypes.
	 * 
	 * @param type
	 *            type
	 * @param superTypes
	 *            supertypes of the original type
	 * @param unnecessaryTransitiveInheritance
	 *            unnecessaryTransitiveInheritance
	 */
	void addTypeToUnnecessaryInheritanceIfSupertypesContainsIt(final Type type,
			final Collection<Type> superTypes,
			final Collection<Type> unnecessaryTransitiveInheritance) {
		if (superTypes.contains(type)) {
			unnecessaryTransitiveInheritance.add(type);
		}
	}
	
	/**
	 * Call the Method getUnnecessaryTransitiveInheritance for every supertype of the supertype.
	 * 
	 * @param supertype
	 *            supertype
	 * @param superTypes
	 *            superTypes of the original type
	 * @param unnecessaryTransitiveInheritance
	 *            unnecessaryTransitiveInheritance
	 */
	void checkUnnecessaryInheritanceInTheSuperTypesOfTheSuperType(final Type supertype,
			final Collection<Type> superTypes,
			final Collection<Type> unnecessaryTransitiveInheritance) {
		for (final Type superTypeOfSuperType : supertype.getSuperTypes()) {
			unnecessaryTransitiveInheritance.addAll(SupertypesCleanInheritanceTask.this
					.getUnnecessaryTransitiveInheritance(superTypeOfSuperType, superTypes));
		}
	}
	
	@Override
	public boolean containsTransitive(final DependencyTask a) {
		return false;
	}
	
	/**
	 * Returns the set of all types, which was found.
	 * 
	 * @return sums
	 */
	public Set<Type> getAllTypesOfModel() {
		return this.allTypesOfModel;
	}
	
	/**
	 * Model.
	 * 
	 * @return Model
	 */
	public Model getModel() {
		return this.model;
	}
	
}
