package de.fhdw.wtf.walker.walker;

import de.fhdw.wtf.common.ast.type.AtomicType;
import de.fhdw.wtf.common.ast.type.BaseType;
import de.fhdw.wtf.common.ast.type.ByNameState;
import de.fhdw.wtf.common.ast.type.ByReferenceState;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.type.CompositeType;
import de.fhdw.wtf.common.ast.type.InvalidState;
import de.fhdw.wtf.common.ast.type.ListType;
import de.fhdw.wtf.common.ast.type.MapType;
import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.ast.type.SumType;
import de.fhdw.wtf.common.ast.type.ThrownType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.type.TypeProxy;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.exception.ast.ASTException;

/**
 * Hilfsklasse für den Umgang mit {@link TypeProxy}´s.
 * 
 */
public final class HelperUtils {
	
	/**
	 * Error-Fehlermeldung, wenn {@link TypeProxy} nicht als State {@link ByReferenceState} hat.
	 * 
	 */
	private static final String ERROR_MESSAGE = "Referencer must run successfully before this task!!! ";
	
	/**
	 * Error-Message: It was a CompositeType.
	 */
	private static final String ERROR_WAS_COMPOSITE_MESSAGE = "It was a CompositeType!";
	
	/**
	 * Error-Message: It was a TypeProxy.
	 */
	private static final String ERROR_WAS_TYPEPROXY_MESSAGE = "It was a TypeProxy!";
	
	/**
	 * Error-Message: It was a BaseType.
	 */
	private static final String ERROR_WAS_BASETYPE_MESSAGE = "It was a BaseType!";
	
	/**
	 * Hilfsklasse für den Umgang mit {@link TypeProxy}´s.
	 */
	private HelperUtils() {
	}
	
	/**
	 * Returns the Type or returns the target of the typeProxy if the Type is a TypeProxy and the state is a
	 * ByReferenceState. If it is another State an ASTException will be thrown.
	 * 
	 * @param type
	 *            Type
	 * @return the type
	 */
	public static Type getReferencedType(final Type type) {
		return type.accept(new TypeVisitorReturn<Type>() {
			
			@Override
			public Type handle(final AtomicType s) {
				return s;
			}
			
			@Override
			public Type handle(final CompositeType c) {
				return c;
			}
			
			@Override
			public Type handle(final TypeProxy s) {
				return s.getState().accept(new TypeProxyStateVisitorReturn<Type>() {
					
					@Override
					public Type handle(final ByNameState byName) {
						throw new ASTException(ERROR_MESSAGE + byName);
					}
					
					@Override
					public Type handle(final InvalidState invalidTypeReference) {
						throw new ASTException(ERROR_MESSAGE + invalidTypeReference.getName());
					}
					
					@Override
					public Type handle(final ByReferenceState byType) {
						return byType.getTarget();
					}
				});
			}
		});
	}
	
	/**
	 * Returns the AtomicType or returns the target of the typeProxy if the Type is a TypeProxy, the state is a
	 * ByReferenceState and the target is also an AtomicType. If it is another State or Type an ASTException will be
	 * thrown.
	 * 
	 * @param type
	 *            Type
	 * @return the type
	 */
	public static AtomicType getReferencedAtomicType(final Type type) {
		return type.accept(new TypeVisitorReturn<AtomicType>() {
			
			@Override
			public AtomicType handle(final AtomicType s) {
				return s;
			}
			
			@Override
			public AtomicType handle(final CompositeType c) {
				throw new ASTException(ERROR_WAS_COMPOSITE_MESSAGE);			}
			
			@Override
			public AtomicType handle(final TypeProxy s) {
				return s.getState().accept(new TypeProxyStateVisitorReturn<AtomicType>() {
					
					@Override
					public AtomicType handle(final ByNameState byName) {
						throw new ASTException(ERROR_MESSAGE + byName);
					}
					
					@Override
					public AtomicType handle(final InvalidState invalidTypeReference) {
						throw new ASTException(ERROR_MESSAGE + invalidTypeReference.getName());
					}
					
					@Override
					public AtomicType handle(final ByReferenceState byType) {
						return getReferencedAtomicType(byType.getTarget());
					}
				});
			}
		});
	}
	
	/**
	 * Returns the ClassType or returns the target of the typeProxy if the Type is a TypeProxy, the state is a
	 * ByReferenceState and the target is also an ClassType. If it is another State or Type an ASTException will be
	 * thrown.
	 * 
	 * @param type
	 *            Type
	 * @return the type
	 */
	public static ClassType getReferencedClassType(final Type type) {
		return type.accept(new TypeVisitorReturn<ClassType>() {
			
			@Override
			public ClassType handle(final AtomicType s) {
				return s.accept(new AtomicTypeVisitorReturn<ClassType>() {
					
					@Override
					public ClassType handle(final BaseType baseType) {
						throw new ASTException(ERROR_WAS_BASETYPE_MESSAGE);
					}
					
					@Override
					public ClassType handle(final ClassType clazz) {
						return clazz;
					}
				});
			}
			
			@Override
			public ClassType handle(final CompositeType c) {
				throw new ASTException(ERROR_WAS_COMPOSITE_MESSAGE + " " + c.toString());
			}
			
			@Override
			public ClassType handle(final TypeProxy s) {
				return s.getState().accept(new TypeProxyStateVisitorReturn<ClassType>() {
					
					@Override
					public ClassType handle(final ByNameState byName) {
						throw new ASTException(ERROR_MESSAGE + byName);
					}
					
					@Override
					public ClassType handle(final InvalidState invalidTypeReference) {
						throw new ASTException(ERROR_MESSAGE + invalidTypeReference.getName());
					}
					
					@Override
					public ClassType handle(final ByReferenceState byType) {
						return getReferencedClassType(byType.getTarget());
					}
				});
			}
		});
	}
	
	/**
	 * Returns the Type of a TypeProxy if the state is a ByReferenceState. If it is another State an ASTException will
	 * be thrown.
	 * 
	 * @param typeProxy
	 *            TypeProxy
	 * @return the type
	 */
	public static Type getTargetType(final TypeProxy typeProxy) {
		return typeProxy.getState().accept(new TypeProxyStateVisitorReturn<Type>() {
			
			@Override
			public Type handle(final ByNameState byName) {
				throw new ASTException(ERROR_MESSAGE + byName);
			}
			
			@Override
			public Type handle(final InvalidState invalidState) {
				throw new ASTException(ERROR_MESSAGE + invalidState.getName());
			}
			
			@Override
			public Type handle(final ByReferenceState byType) {
				return byType.getTarget();
			}
		});
	}
	
	/**
	 * Returns the Type of a TypeProxy if the state is a ByReferenceState. If it is another State null is returned.
	 * 
	 * @param typeProxy
	 *            TypeProxy
	 * @return the type or null
	 */
	public static Type tryGetTargetType(final TypeProxy typeProxy) {
		return typeProxy.getState().accept(new TypeProxyStateVisitorReturn<Type>() {
			
			@Override
			public Type handle(final ByNameState byName) {
				return null;
			}
			
			@Override
			public Type handle(final InvalidState invalidState) {
				return null;
			}
			
			@Override
			public Type handle(final ByReferenceState byType) {
				return byType.getTarget();
			}
		});
	}
	
	/**
	 * Returns the AtomicType of a TypeProxy if the state is a ByReferenceState. If it is another State or not an
	 * AtomicType, an ASTException will be thrown.
	 * 
	 * @param typeProxy
	 *            TypeProxy
	 * @return the type
	 */
	public static AtomicType getTargetAtomicType(final TypeProxy typeProxy) {
		return typeProxy.getState().accept(new TypeProxyStateVisitorReturn<AtomicType>() {
			
			@Override
			public AtomicType handle(final ByNameState byName) {
				throw new ASTException(ERROR_MESSAGE + byName);
			}
			
			@Override
			public AtomicType handle(final InvalidState invalidState) {
				throw new ASTException(ERROR_MESSAGE + invalidState.getName());
			}
			
			@Override
			public AtomicType handle(final ByReferenceState byType) {
				return byType.getTarget().accept(new TypeVisitorReturn<AtomicType>() {
					
					@Override
					public AtomicType handle(final AtomicType s) {
						return s;
					}
					
					@Override
					public AtomicType handle(final CompositeType c) {
						throw new ASTException(ERROR_WAS_COMPOSITE_MESSAGE);
					}
					
					@Override
					public AtomicType handle(final TypeProxy s) {
						throw new ASTException(ERROR_WAS_TYPEPROXY_MESSAGE);
					}
				});
			}
		});
	}
	
	/**
	 * Returns the ClassType of a TypeProxy if the state is a ByReferenceState. If it is another State or not an
	 * ClassType, an ASTException will be thrown.
	 * 
	 * @param typeProxy
	 *            TypeProxy
	 * @return the type
	 */
	public static ClassType getTargetClassType(final TypeProxy typeProxy) {
		return typeProxy.getState().accept(new TypeProxyStateVisitorReturn<ClassType>() {
			
			@Override
			public ClassType handle(final ByNameState byName) {
				throw new ASTException(ERROR_MESSAGE + byName);
			}
			
			@Override
			public ClassType handle(final InvalidState invalidState) {
				throw new ASTException(ERROR_MESSAGE + invalidState.getName());
			}
			
			@Override
			public ClassType handle(final ByReferenceState byType) {
				return byType.getTarget().accept(new TypeVisitorReturn<ClassType>() {
					
					@Override
					public ClassType handle(final AtomicType s) {
						return s.accept(new AtomicTypeVisitorReturn<ClassType>() {
							
							@Override
							public ClassType handle(final BaseType baseType) {
								throw new ASTException(ERROR_WAS_BASETYPE_MESSAGE);
							}
							
							@Override
							public ClassType handle(final ClassType clazz) {
								return clazz;
							}
						});
					}
					
					@Override
					public ClassType handle(final CompositeType c) {
						throw new ASTException(ERROR_WAS_COMPOSITE_MESSAGE);
					}
					
					@Override
					public ClassType handle(final TypeProxy s) {
						throw new ASTException(ERROR_WAS_TYPEPROXY_MESSAGE);
					}
				});
			}
		});
	}
	
	/**
	 * Returns the ByReferenceState of a TypeProxy if the state is a ByReferenceState. If it is another State an
	 * ASTException will be thrown.
	 * 
	 * @param typeProxy
	 *            TypeProxy
	 * @return the ByReferenceState
	 */
	public static ByReferenceState getByReferenceState(final TypeProxy typeProxy) {
		return typeProxy.getState().accept(new TypeProxyStateVisitorReturn<ByReferenceState>() {
			
			@Override
			public ByReferenceState handle(final ByNameState byName) {
				throw new ASTException(ERROR_MESSAGE + byName);
			}
			
			@Override
			public ByReferenceState handle(final InvalidState invalidState) {
				throw new ASTException(ERROR_MESSAGE + invalidState.getName());
			}
			
			@Override
			public ByReferenceState handle(final ByReferenceState byType) {
				return byType;
			}
		});
	}
}
