package de.fhdw.wtf.walker.walker;

import java.util.HashSet;
import java.util.Set;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.Group;
import de.fhdw.wtf.common.ast.GroupElement;
import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.Operation;
import de.fhdw.wtf.common.ast.type.AtomicType;
import de.fhdw.wtf.common.ast.type.BaseType;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.type.CompositeType;
import de.fhdw.wtf.common.ast.type.MapType;
import de.fhdw.wtf.common.ast.type.ProductElementType;
import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.ast.type.SumType;
import de.fhdw.wtf.common.ast.type.ThrownType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.type.TypeProxy;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.GroupElementExceptionVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorException;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.DependencyTask;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.common.task.result.OKTaskResult;

/**
 * Visits each referenced type in the model once.
 */
@SuppressWarnings("synthetic-access")
public abstract class SimpleWalkerTaskForTypes extends WalkerTaskForTypes {
	
	/**
	 * Abstract syntax tree model traversed by this task.
	 */
	private final Model m;
	
	/**
	 * All types referenced by the model.
	 */
	private final Set<Type> foundTypes;
	
	/**
	 * Flag for using Prototype or not.
	 */
	private final boolean usePrototype;
	
	/**
	 * Creates a {@link SimpleWalkerTaskForTypes} walker. The walker will not visit prototypes.
	 * 
	 * @param m
	 *            The model to traverse.
	 * @param taskmanager
	 *            The task manager to register with.
	 */
	public SimpleWalkerTaskForTypes(final Model m, final TaskExecutor taskmanager) {
		this(m, taskmanager, false);
	}
	
	/**
	 * Creates a {@link SimpleWalkerTaskForTypes} walker. Allows to specify whether prototypes should be visited or not.
	 * Visit prototypes only after the PrototypesTask!
	 * 
	 * @param m
	 *            The model to traverse.
	 * @param taskmanager
	 *            The task manager to register with.
	 * @param usePrototype
	 *            If true, this walker will visit prototypes.
	 */
	public SimpleWalkerTaskForTypes(final Model m, final TaskExecutor taskmanager, final boolean usePrototype) {
		super(taskmanager);
		this.m = m;
		this.foundTypes = new HashSet<>();
		this.usePrototype = usePrototype;
	}
	
	@Override
	public de.fhdw.wtf.common.task.result.TaskResult doWork() {
		try {
			this.beginTask();
			this.walk();
			this.finalizeTask();
		} catch (final TaskException e) {
			return new de.fhdw.wtf.common.task.result.ExceptionalTaskResult(e);
		}
		return new OKTaskResult();
	}
	
	/**
	 * Traverses the tree and collects all referenced types.
	 * 
	 * @throws TaskException
	 *             when the implementing task throws unfixable errors
	 */
	private void walk() throws TaskException {
		this.processCompositeType(this.m.getAnything());
		for (final Group g : this.m.getGroups()) {
			this.processGroupElements(g);
		}
		
		for (final Type t : this.foundTypes) {
			this.handleType(t);
		}
	}
	
	/**
	 * Processes a group by descending into its members.
	 * 
	 * @param g
	 *            The group to process.
	 * @throws TaskException
	 *             when the implementing task throws unfixable errors
	 */
	private void processGroupElements(final Group g) throws TaskException {
		for (final GroupElement element : g.getGroupElements()) {
			element.accept(new GroupElementExceptionVisitor<TaskException>() {
				@Override
				public void handle(final Group g2) throws TaskException {
					SimpleWalkerTaskForTypes.this.processGroupElements(g2);
				}
				
				@Override
				public void handle(final ClassType clss) throws TaskException {
					SimpleWalkerTaskForTypes.this.processClassType(clss);
				}
			});
		}
	}
	
	/**
	 * Processes a class type.
	 * 
	 * @param clss
	 *            The class type.
	 * @throws TaskException
	 *             when the implementing task throws unfixable errors
	 */
	private void processClassType(final ClassType clss) throws TaskException {
		ClassType useClass = clss;
		if (this.isUsePrototype()) {
			useClass = useClass.getPrototype();
		}
		if (!this.foundTypes.contains(useClass)) {
			this.foundTypes.add(useClass);
			for (final Attribute a : useClass.getAttributes()) {
				SimpleWalkerTaskForTypes.this.processTypesInAttribute(a);
			}
			for (final Operation o : useClass.getOperations()) {
				SimpleWalkerTaskForTypes.this.processTypesInOperation(o);
			}
		}
	}
	
	/**
	 * Processes a composite type.
	 * 
	 * @param type
	 *            The composite type.
	 * @throws TaskException
	 *             when the implementing task throws unfixable errors
	 */
	private void processCompositeType(final CompositeType type) throws TaskException {
		if (this.isUsePrototype()) {
			final Type t = type.getPrototype();
			if (type != t) { // NOPMD - must compare references here, not objects
				this.processType(t);
				return;
			}
		}
		if (!this.foundTypes.contains(type)) {
			this.foundTypes.add(type);
			type.accept(new SearchTypesInTypeCompositeVisitor());
		}
	}
	
	/**
	 * Processes types referenced by some operation.
	 * 
	 * @param o
	 *            The operation to analyze.
	 * @throws TaskException
	 *             if collecting the types fails for some reason
	 */
	private void processTypesInOperation(final Operation o) throws TaskException {
		this.processType(o.getReturnType());
		this.processType(o.getParameters());
	}
	
	/**
	 * Processes types referenced by some attribute.
	 * 
	 * @param a
	 *            The attribute to analyze.
	 * @throws TaskException
	 *             if collecting the types fails for some reason
	 */
	private void processTypesInAttribute(final Attribute a) throws TaskException {
		this.processType(a.getAttrType());
	}
	
	/**
	 * Processes types referenced by some type.
	 * 
	 * @param t
	 *            The type to analyze
	 * @throws TaskException
	 *             if collecting the types fails for some reason
	 */
	private void processType(final Type t) throws TaskException {
		t.accept(new SearchTypesInTypeTypeVisitor());
	}
	
	/**
	 * Visits types referenced by some type.
	 */
	private final class SearchTypesInTypeTypeVisitor implements TypeVisitorException<TaskException> {
		
		/**
		 * Creates a {@link SearchTypesInTypeTypeVisitor} object.
		 */
		protected SearchTypesInTypeTypeVisitor() {
			super();
		}
		
		@Override
		public void handle(final AtomicType s) throws TaskException {
			s.accept(new AtomicTypeVisitorException<TaskException>() {
				@Override
				public void handle(final BaseType baseType) throws TaskException {
					SimpleWalkerTaskForTypes.this.getFoundTypes().add(baseType);
				}
				
				@Override
				public void handle(final ClassType clazz) throws TaskException {
					SimpleWalkerTaskForTypes.this.processClassType(clazz);
				}
			});
		}
		
		@Override
		public void handle(final CompositeType c) throws TaskException {
			SimpleWalkerTaskForTypes.this.processCompositeType(c);
		}
		
		@Override
		public void handle(final TypeProxy typeProxy) throws TaskException {
			final Type targetType = HelperUtils.tryGetTargetType(typeProxy);
			if (targetType != null) {
				targetType.accept(this);
			}
		}
	}
	
	/**
	 * Visits types referenced by some composite type.
	 */
	private final class SearchTypesInTypeCompositeVisitor implements CompositeTypeVisitorException<TaskException> {
		
		/**
		 * Creates a {@link SearchTypesInTypeCompositeVisitor} object.
		 */
		protected SearchTypesInTypeCompositeVisitor() {
			super();
		}
		
		@Override
		public void handle(final de.fhdw.wtf.common.ast.type.ListType list) throws TaskException {
			SimpleWalkerTaskForTypes.this.processType(list.getOf());
		}
		
		@Override
		public void handle(final MapType map) throws TaskException {
			SimpleWalkerTaskForTypes.this.processType(map.getKey());
			SimpleWalkerTaskForTypes.this.processType(map.getOf());
		}
		
		@Override
		public void handle(final ProductType product) throws TaskException {
			final ProductType abstractProductPrototype = product.getAbstractPrototype();
			if (abstractProductPrototype != null && !product.equals(abstractProductPrototype)) {
				SimpleWalkerTaskForTypes.this.processCompositeType(abstractProductPrototype);
			}
			for (final ProductElementType current : product.getElements()) {
				SimpleWalkerTaskForTypes.this.processType(current.getType());
			}
		}
		
		@Override
		public void handle(final SumType sum) throws TaskException {
			for (final Type current : sum.getElements()) {
				SimpleWalkerTaskForTypes.this.processType(current);
			}
		}
		
		@Override
		public void handle(final ThrownType thrownType) throws TaskException {
			SimpleWalkerTaskForTypes.this.processType(thrownType.getUnderlyingType());
		}
	}
	
	@Override
	public boolean containsTransitive(final DependencyTask a) {
		return false;
	}
	
	/**
	 * Get the attribute UsePrototype.
	 * 
	 * @return UsePrototype
	 */
	public boolean isUsePrototype() {
		return this.usePrototype;
	}
	
	/**
	 * Returns all types referenced by the model.
	 * 
	 * @return All types referenced by the model.
	 */
	protected Set<Type> getFoundTypes() {
		return this.foundTypes;
	}
	
	/**
	 * Returns the underlying model.
	 * 
	 * @return The underlying model.
	 */
	public Model getModel() {
		return this.m;
	}
	
}
